﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace LinqToNothing {
	class Nothing {
		public Nothing Where(Predicate<int> p) {
			Console.WriteLine("Nothing.Where called.");
			return this;
		}

		public override string ToString() {
			return "Really nothing!";
		}
	}

	class NothingIntArg {
		public NothingIntArg Where(Func<int, int> f) {
			Console.WriteLine("NothingIntArg.Where called.");
			return this;
		}

		public override string ToString() {
			return "NothingIntArg's Really nothing!";
		}
	}

	class NothingInt5 {
		public int Where(Predicate<int> p) {
			Console.WriteLine("NothingInt5.Where called.");
			return 5;
		}

		public override string ToString() {
			return "5 is really nothing!";
		}
	}

	class Program {
		static void Main(string[] args) {
			var q1 = from x in new Nothing() where x > 5 select x;
			Console.WriteLine(q1);
			Console.WriteLine();

			var q1ia = from x in new NothingIntArg() where x + 5 select x;
			Console.WriteLine(q1ia);
			Console.WriteLine();

			var qi5 = from x in new NothingInt5() where x > 5 select x;
			Console.WriteLine(qi5);
			Console.WriteLine();

			var q2 = from x in new Nothing() where x > 5 where x < 10 select x;
			Console.WriteLine(q2);
			Console.WriteLine();

			// Following query will not compile because .Where(x => x > 5) on NothingInt5 returns an int 
			// and no Where method exists on int (to evaluate the following .Where(x => x < 10).
			// Error: Could not find an implementation of the query pattern for source type 'int'. 'Where' not found.
			//		var q2i5 = from x in new NothingInt5() where x > 5 where x < 10 select x;

			// Following query will not compile because of a missing Select method in Nothing class.
			// Error: Could not find an implementation of the query pattern for source type 'LinqToNothing.Nothing'. 'Select' not found.
			//		var q3 = from x in new Nothing() where x > 5 where x < 10 select 10;

			var qa = from x in new NothingAdvanced() where x > 5 where x < 10 select 10;
			Console.WriteLine(qa);
			Console.WriteLine();
		}
	}

	class NothingAdvanced {
		public NothingAdvanced Where(Predicate<int> p) {
			Console.WriteLine("NothingAdvanced.Where called.");
			return this;
		}

		public int Select(Func<int, int> f) {
			Console.WriteLine("NothingAdvanced.Select called.");
			return 42;
		}

		public override string ToString() {
			return "Really nothing, but advanced!";
		}
	}
}
