﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace LinqToObjectsInside {

	class DataSource : IEnumerable<int> {
		bool throwAtEnd;

		public DataSource(bool throwAtEnd) {
			this.throwAtEnd = throwAtEnd;
		}

		public IEnumerator<int> GetEnumerator() {
			Console.WriteLine("DataSource: yield return 1");
			yield return 1;
			Console.WriteLine("DataSource: yield return 2");
			yield return 2;
			Console.WriteLine("DataSource: yield return 5");
			yield return 5;
			Console.WriteLine("DataSource: yield return 11");
			yield return 11;
			Console.WriteLine("DataSource: yield return 6");
			yield return 6;
			Console.WriteLine("DataSource: yield return 7");
			yield return 7;
			if (throwAtEnd) {
				throw new Exception("DataSource: Done.");
			} else {
				Console.WriteLine("DataSource: Done.");
			}
		}

		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() {
			return GetEnumerator();
		}
	}

	class Program {
		static bool p1(int i) {
			Console.WriteLine("  where p1({0} > 2?)", i);
			return i > 2;
		}

		static bool p2(int i) {
			Console.WriteLine("    where p2({0} < 10?)", i);
			return i < 10;
		}

		static void Main(string[] args) {

			Console.WriteLine("before query ...");
			var q = from i in new DataSource(false) where p1(i) where p2(i) select i + 100;

			Console.WriteLine("before foreach ...");
			foreach (var i in q) {
				Console.WriteLine("*** got {0}", i);
			}

			Console.WriteLine();
			Console.WriteLine("before query ...");
			q = from i in new DataSource(false) where p1(i) where p2(i) select i + 100;

			Console.WriteLine("before foreach ...");
			foreach (var i in q) {
				Console.WriteLine("*** got {0}", i);
				break;	// !!!
			}

			Console.WriteLine();
			Console.WriteLine("=====================================");

			/**/

			Console.WriteLine();	// additional OrderBy
			Console.WriteLine("before query ...");
			q = from i in new DataSource(false) where p1(i) where p2(i) orderby i select i + 100;

			Console.WriteLine("before foreach ...");
			foreach (var i in q) {
				Console.WriteLine("*** got {0}", i);
			}

			Console.WriteLine();
			Console.WriteLine("before query ...");
			q = from i in new DataSource(false) where p1(i) where p2(i) orderby i select i + 100;

			Console.WriteLine("before foreach ...");
			foreach (var i in q) {
				Console.WriteLine("*** got {0}", i);
				break;  // !!!
			}

			Console.WriteLine();
			Console.WriteLine("+++++++++++++++++++++++++++++++++++++");

			/**/
			/**/

			Console.WriteLine();    // OrderBy now first in query
			Console.WriteLine("before query ...");
			q = from i in new DataSource(false) orderby i where p1(i) where p2(i) select i + 100;

			Console.WriteLine("before foreach ...");
			foreach (var i in q) {
				Console.WriteLine("*** got {0}", i);
			}

			/**/
			/**/

			// NOTE: View stack trace of exception thrown from DataSource to see LINQ to Objects implementation ...
			Console.WriteLine();
			Console.WriteLine("before query ...");
			q = from i in new DataSource(true) where p1(i) where p2(i) orderby i select i * 2;

			Console.WriteLine("before foreach ...");
			foreach (var i in q) {
				Console.WriteLine("*** got {0}", i);
			}

			/**/
		}
	}
}
