﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace LinqToObjectsQueryReuse {
	class Program {
		static IEnumerable<int> GetSortedGreaterThan3(IEnumerable<int> source) {
			return from i in source where i > 3 orderby i select i;
		}

		static IEnumerable<int> GetMultipliedBy2(IEnumerable<int> source) {
			return from i in source select i * 2;
		}

		static IEnumerable<int> GetLessThan10(IEnumerable<int> source) {
			return from i in source where i < 10 select i;
		}


		static void Main(string[] args) {
			List<int> numbers = new List<int> { 1, 2, 4, 5, 6, 10, 20, 30 };

			var q1 = GetSortedGreaterThan3(numbers);
			Print(q1); Console.WriteLine();

			Console.WriteLine("Adding item ...");
			numbers.Add(7);
			Print(q1); Console.WriteLine();

			var q2 = GetMultipliedBy2(q1);

			Console.WriteLine("Removing items ...");
			numbers.RemoveAll(x => x < 5);
			Print(q1);
			Print(q2); Console.WriteLine();

			var q3 = GetLessThan10(q1);
			var q4 = GetLessThan10(q2);

			Console.WriteLine("Adding range of items ...");
			numbers.AddRange(Enumerable.Range(1, 5));
			Print(q1);
			Print(q2);
			Print(q3);
			Print(q4); Console.WriteLine();

			Console.WriteLine("Replacing whole source ...");
			numbers = new List<int> { 100, 200, 300 };
			Print(q1);
			q1 = GetSortedGreaterThan3(numbers);
			Print(q1);
		}

		static void Print(IEnumerable<int> data) {
			foreach (int i in data) {
				Console.Write("{0} ", i);
			}
			Console.WriteLine();
		}
	}
}
